﻿// MP4 data processor
//    for moraMP4fix_for_NW-A860
//
// Copyright(C) 2014 Sota. All rights reserved.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace moraMP4fix_for_NW_A860
{
    class Mp4Processing
    {
        int moovBoxOffset;
        int udtaBoxOffset;
        int moovSubtraction;
        int udtaSubtraction;

        /// <summary>
        /// Constructor
        /// </summary>
        public Mp4Processing()
        {
            moovBoxOffset = -1;
            udtaBoxOffset = -1;
            moovSubtraction = 0;
            udtaSubtraction = 0;
        }

        /// <summary>
        /// Check whether file is MP4 file
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="size">MP4 data size</param>
        /// <retunrs>true=MP4 file</retunrs>
        public bool CheckMp4File(byte[] mp4Data, int size)
        {
            bool mp4 = false;
            if (size > 8)
            {
                string atom = GetAtom(mp4Data, 0);
                if (atom == "ftyp")
                {
                    mp4 = true;
                }
            }
            return mp4;
        }

        /// <summary>
        /// Modify MP4 data
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="size">MP4 data size</param>
        /// <param name="clip">Clip manager</param>
        /// <param name="gui">GUI</param>
        public void ModifyMp4Data(byte[] mp4Data, int size, ClipManager clip, Form1 gui)
        {
            TrackManager track = new TrackManager();

            AnalyzeMp4Data(mp4Data, 0, size, "", clip, track, gui);

            UpdateMoovBoxSize(mp4Data);
            UpdateUdtaBoxSize(mp4Data);

#if false
            for (int i = 0; i < track.count(); i++)
            {
                TrackManager.TrackData trackData = track.get(i);
                gui.showMessage(string.Format("stco offset=0x{0:x}", trackData.stcoOffset));
            }
#endif

            UpdateChunkOffset(mp4Data, track);

        }

        /// <summary>
        /// Analyze and modify MP4 data
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="startOffset">Analyze start offset</param>
        /// <param name="size">MP4 data size</param>
        /// <param name="parent">Parent atom string</param>
        /// <param name="clip">Clip manager</param>
        /// <param name="track">Track manager</param>
        /// <param name="gui">GUI</param>
        private void AnalyzeMp4Data(byte[] mp4Data, int startOffset, int size, string parent, ClipManager clip, TrackManager track, Form1 gui)
        {
            int offset = 0;

            while (offset < size)
            {
                int boxSize = GetBoxSize(mp4Data, startOffset + offset);
                string atom = GetAtom(mp4Data, startOffset + offset);

#if false
                gui.showMessage(atom);
#endif

                if ((parent == "") && (atom == "moov"))
                {
                    moovBoxOffset = startOffset + offset;
                    AnalyzeMp4Data(mp4Data, startOffset + offset + 8, boxSize - 8, atom, clip, track, gui);
                }
                else if ((parent == "moov") && (atom == "udta"))
                {
                    udtaBoxOffset = startOffset + offset;
                    AnalyzeMp4Data(mp4Data, startOffset + offset + 8, boxSize - 8, atom, clip, track, gui);
                }
                else if (((parent == "moov") && (atom == "trak")) ||
                         ((parent == "trak") && (atom == "mdia")) ||
                         ((parent == "mdia") && (atom == "minf")) ||
                         ((parent == "minf") && (atom == "stbl")))
                {
                    AnalyzeMp4Data(mp4Data, startOffset + offset + 8, boxSize - 8, atom, clip, track, gui);
                }
                else if ((parent == "moov") && (atom == "meta"))
                {
                    moovSubtraction += boxSize;
                    clip.AddClipData(startOffset + offset, boxSize);
                }
                else if ((parent == "udta") &&
                            ((atom == "albm") ||
                             (atom == "cprt") ||
                             (atom == "dscp") ||
                             (atom == "gnre") ||
                             (atom == "perf") ||
                             (atom == "titl") ||
                             (atom == "yrrc")))
                {
                    udtaSubtraction += boxSize;
                    moovSubtraction += boxSize;
                    clip.AddClipData(startOffset + offset, boxSize);
                }
                else if ((parent == "stbl") && (atom == "stco"))
                {
                    track.AddTrackData(startOffset + offset);
                }

                offset += boxSize;
            }
        }

        /// <summary>
        /// Get MP4 box size
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="offset">Offset of box</param>
        /// <retunrs>Box size</retunrs>
        private int GetBoxSize(byte[] mp4Data, int offset)
        {
            int size = GetInt32Data(mp4Data, offset);
            return size;
        }

        /// <summary>
        /// Set MP4 box size
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="offset">Offset of box</param>
        /// <param name="boxSize">Box size</param>
        private void SetBoxSize(byte[] mp4Data, int offset, int boxSize)
        {
            SetInt32Data(mp4Data, offset, boxSize);
        }

        /// <summary>
        /// get atom
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="offset">Offset of box</param>
        /// <retunrs>Atom string</retunrs>
        private string GetAtom(byte[] mp4Data, int offset)
        {
            byte[] atom = new byte[4];
            atom[0] = mp4Data[offset + 0 + 4];
            atom[1] = mp4Data[offset + 1 + 4];
            atom[2] = mp4Data[offset + 2 + 4];
            atom[3] = mp4Data[offset + 3 + 4];
            Encoding utf8enc = Encoding.GetEncoding("utf-8");
            string text = utf8enc.GetString(atom);
            return text;
        }

        /// <summary>
        /// Get Int32 value (Big endian)
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="offset">Offset</param>
        /// <retunrs>Int32 value</retunrs>
        private int GetInt32Data(byte[] mp4Data, int offset)
        {
            byte[] sub = new byte[4];
            sub[0] = mp4Data[offset + 3];
            sub[1] = mp4Data[offset + 2];
            sub[2] = mp4Data[offset + 1];
            sub[3] = mp4Data[offset + 0];
            int size = BitConverter.ToInt32(sub, 0);
            return size;
        }

        /// <summary>
        /// Set Int32 value (Big endian)
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="offset">Offset</param>
        /// <param name="value">Int32 value</param>
        private void SetInt32Data(byte[] mp4Data, int offset, int value)
        {
            byte[] size = BitConverter.GetBytes(value);
            mp4Data[offset + 3] = size[0];
            mp4Data[offset + 2] = size[1];
            mp4Data[offset + 1] = size[2];
            mp4Data[offset + 0] = size[3];
        }

        /// <summary>
        /// Update moov box size
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        private void UpdateMoovBoxSize(byte[] mp4Data)
        {
            if ((moovBoxOffset != -1) && (moovSubtraction > 0))
            {
                int moovSize = GetBoxSize(mp4Data, moovBoxOffset);
                moovSize -= moovSubtraction;
                SetBoxSize(mp4Data, moovBoxOffset, moovSize);
            }
        }

        /// <summary>
        /// Update udta (moov.udta) box size
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        private void UpdateUdtaBoxSize(byte[] mp4Data)
        {
            if ((udtaBoxOffset != -1) && (udtaSubtraction > 0))
            {
                int udtaSize = GetBoxSize(mp4Data, udtaBoxOffset);
                udtaSize -= udtaSubtraction;
                SetBoxSize(mp4Data, udtaBoxOffset, udtaSize);
            }
        }

        /// <summary>
        /// Update chunk offset array
        /// </summary>
        /// <param name="mp4Data">MP4 data top</param>
        /// <param name="track">Track manager</param>
        private void UpdateChunkOffset(byte[] mp4Data, TrackManager track)
        {
            if (moovSubtraction > 0)
            {
                for (int i = 0; i < track.GetTrackDataCount(); i++)
                {
                    TrackManager.TrackData trackData = track.GetTrackData(i);
                    int stcoSize = GetBoxSize(mp4Data, trackData.stcoOffset);
                    if (stcoSize > 16)
                    {
                        int entryCount = GetInt32Data(mp4Data, trackData.stcoOffset + 12);
                        for (int entry = 0; entry < entryCount; entry++)
                        {
                            int chunkOffset = GetInt32Data(mp4Data, trackData.stcoOffset + 16 + entry * 4);
                            if (chunkOffset > moovBoxOffset)
                            {
                                chunkOffset -= moovSubtraction;
                                SetInt32Data(mp4Data, trackData.stcoOffset + 16 + entry * 4, chunkOffset);
                            }
                        }
                    }
                }
            }
        }

    }
}
